/*
*
* Program to combine files which were split
* with Windows Commander
*
* BSD-Unix (MacOS X 10.1.x) by
* (C) 2002 Andreas Kromke
*
* Adapted to Win32 Console by Christian Ghisler
*
* Contents of a CRC file:
*
* filename=<Name of original file>
* size=<Size of original file, in bytes, decimal>
* crc32=<32-Bit-checksum, hex, same as used in zlib>
*
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
//#include <unistd.h>
#include <io.h>
#include <fcntl.h>
#include <errno.h>
//#include <termios.h>
#include <sys/stat.h>
#include "zlib.h"

#define BUFSIZE 131072

#define min(a,b) (((a) < (b)) ? (a) : (b))

struct CRCInfo
{
	char FileName[256];
	unsigned long Size;
	unsigned long crc;
};

static int LoadFile(const char *FileName, char **pData, size_t *pFileSize);
static int AnalyzeCRCFile(char *pData, size_t FileSize, CRCInfo *pInfo);
static int AppendFile(int fd, const char *FileName, size_t *plen, unsigned long *pCrc);

int main (int argc, const char * argv[])
{
	struct CRCInfo Info;
	char *buf;
	size_t len;
	int ret;
	int fd;
	char *s;
	char Path[256];
	register int i;
	unsigned long crc;


	puts("");
	puts("Combine tool for files split with Windows Commander");
	puts("(C) Andreas Kromke 2002");
	puts("");

	if	(argc != 2)
	{
		puts("Syntax: CombineCRC Datei[.crc]");
		return(1);
	}

	// get path of input files

	if	(strlen(argv[1]) > 255)
	{
		fprintf(stderr, "Path too long: %s\n", argv[1]);
		return(1);
	}
	strcpy(Path, argv[1]);

	s = Path + strlen(Path);
	if	((s - Path > 4) && (!stricmp(s - 4, ".crc")))
	{
		s -= 4;
		*s = '\0';			// Cut extension
	}

	// open CRC file
	ret = LoadFile(argv[1], &buf, &len);
	if	(ret)
		return(ret);

	// analyze CRC file
	ret = AnalyzeCRCFile(buf, len, &Info);
	if	(ret)
		return(ret);

	// free CRC file
	free(buf);

	// Create target file, only if it doesn't exist yet
	fd = open(Info.FileName, O_WRONLY | O_CREAT | O_EXCL | O_BINARY, 0777);
	if	(fd == -1)
	{
		fprintf(stderr, "Cannot create file %s. Error %d (%s)\n", Info.FileName, errno, strerror(errno));
		return(errno);
	}

	// Loop

	i = 1;
	len = 0;
	crc = crc32(0L, Z_NULL, 0); 
	do
	{
		// create filename
		sprintf(s, ".%03d", i);
		ret = AppendFile(fd, Path, &len, &crc);
		i++;
	}
	while(!ret);

	close(fd);

	if	(len != Info.Size)
	{
		fprintf(stderr, "File length is %ld instead of the expected %ld\n", len, Info.Size);
		return(errno);
	}

	if	(crc != Info.crc)
	{
		fprintf(stderr, "CRC32-Error! Checksum is 0x%08lx instead of the expected 0x%08lx\n", crc, Info.crc);
		return(1);
	}

	printf("\nCombine of %s successful,\nlength and checksum OK.\n", Info.FileName);

	return(0);
}


/*************************************************************************
*
* Analyses the CRC file
*
*************************************************************************/

static int AnalyzeCRCFile(char *pData, size_t FileSize, CRCInfo *pInfo)
{
	char *s,*t;

	// Line "filename="

	if	((FileSize < 9) || strncmp(pData, "filename=", 9))
	{
		fprintf(stderr, "Syntax error in CRC file (filename=)");
		return(1);
	}

	pData += 9;
	FileSize -= 9;
	s = (char *) memchr(pData, '\r', FileSize);
	if	(!s)
		s = (char *) memchr(pData, '\n', FileSize);
	if	(!s)
		s = pData + FileSize;

	if	(s - pData > 255)
	{
		fprintf(stderr, "Syntax error in CRC file (filename=)");
		return(1);
	}

	memcpy(pInfo->FileName, pData, s - pData);
	pInfo->FileName[s - pData] = '\0';

	FileSize -= (s - pData);
	pData = s;
	while((*pData == '\r' || *pData == '\n') && (FileSize))
	{
		pData++;
		FileSize--;
	}

	// Line "size="
	if	((FileSize < 5) || strncmp(pData, "size=", 5))
	{
		fprintf(stderr, "Syntax error in CRC file (size=)");
		return(1);
	}

	pData += 5;
	FileSize -= 5;
	s = (char *) memchr(pData, '\r', FileSize);
	if	(!s)
		s = (char *) memchr(pData, '\n', FileSize);
	if	(!s)
		s = pData + FileSize;

	pInfo->Size = strtoul(pData, &t, 10);
	if	(s != t)
	{
		fprintf(stderr, "Syntax error in CRC file (size=)");
		return(1);
	}

	FileSize -= (s - pData);
	pData = s;
	while((*pData == '\r' || *pData == '\n') && (FileSize))
	{
		pData++;
		FileSize--;
	}

	// Line "crc="
	if	((FileSize < 6) || strncmp(pData, "crc32=", 6))
	{
		fprintf(stderr, "Syntax error in CRC file (crc32=)");
		return(1);
	}

	pData += 6;
	FileSize -= 6;
	s = (char *) memchr(pData, '\r', FileSize);
	if	(!s)
		s = (char *) memchr(pData, '\n', FileSize);
	if	(!s)
		s = pData + FileSize;

	pInfo->crc = strtoul(pData, &t, 16);
	if	(s != t)
	{
		fprintf(stderr, "Syntax error in CRC file (crc32=)");
		return(1);
	}

	FileSize -= (s - pData);
	pData = s;
	while((*pData == '\r' || *pData == '\n') && (FileSize))
	{
		pData++;
		FileSize--;
	}

	return(0);
}


/*************************************************************************
*
* Reads a file into memory
*
*************************************************************************/

static int LoadFile(const char *FileName, char **pData, size_t *pFileSize)
{
    int fileDescriptor;
	size_t numBytesRead;
	struct stat sd;


	*pData = NULL;
	*pFileSize = 0;

	fileDescriptor = open(FileName, O_RDONLY | O_BINARY);
	if	(fileDescriptor == -1)
	{
		fprintf(stderr, "Cannot open file %s. Error %d (%s)\n", FileName, errno, strerror(errno));
		return(2);
	}

	// Dateigre ermitteln

	if	(-1 == fstat(fileDescriptor, &sd))
	{
		close(fileDescriptor);
		fprintf(stderr, "Cannot check file %s. Fehler %d (%s)\n", FileName, errno, strerror(errno));
		return(2);
	}

	*pFileSize = (size_t) sd.st_size;
	*pData = (char *) malloc(*pFileSize);
	if	(!*pData)
	{
		close(fileDescriptor);
		fprintf(stderr, "Zuwenig Speicher\n");
		return(2);
	}

	// Read file, max. <sd.st_size> Bytes
	// and close

	numBytesRead = read(fileDescriptor, *pData, *pFileSize);
    close(fileDescriptor);

	if	(numBytesRead != *pFileSize)
	{
		free(*pData);
		*pData = NULL;
		*pFileSize = 0;
		fprintf(stderr, "Read error\n");
		return(2);
	}

	return(0);
}


/*************************************************************************
*
* Append file
*
*************************************************************************/

static int AppendFile(int fd, const char *FileName, size_t *pLen, unsigned long *pCrc)
{
    int fileDescriptor;
	size_t ActSize,numBytesRead,FileSize;
	struct stat sd;
	char buf[BUFSIZE];


	fileDescriptor = open(FileName, O_RDONLY | O_BINARY);
	if	(fileDescriptor == -1)
	{
		if	(errno != 2)
			fprintf(stderr, "Cannot open file %s. Error %d (%s)\n", FileName, errno, strerror(errno));
		return(errno);
	}

	// Get file size

	if	(-1 == fstat(fileDescriptor, &sd))
	{
		close(fileDescriptor);
		fprintf(stderr, "Cannot check file %s. Error %d (%s)\n", FileName, errno, strerror(errno));
		return(errno);
	}

	FileSize = (size_t) sd.st_size;
	printf("File %s, Size %ld\n", FileName, FileSize);

	// read/write

	while(FileSize)
	{
		// Get block size
		ActSize = min(FileSize, BUFSIZE);
		// read
		numBytesRead = read(fileDescriptor, buf, ActSize);
		if	(numBytesRead != ActSize)
		{
			fprintf(stderr, "Read error in file %s.\n", FileName);
			close(fileDescriptor);
			return(2);
		}
		// schreiben
		numBytesRead = write(fd, buf, ActSize);
		if	(numBytesRead != ActSize)
		{
			fprintf(stderr, "Error writing to target file!\n");
			close(fileDescriptor);
			return(2);
		}
		*pLen += ActSize;
		FileSize -= ActSize;
		*pCrc = crc32(*pCrc, (unsigned char *) buf, ActSize);
		printf("%8ld kBytes remaining.\r", FileSize >> 10);
		fflush(stdout);
	}

	close(fileDescriptor);

	printf("                              \r");

	return(0);
}
